import traceback
from typing import Literal

import celery
from django.conf import settings
from django.utils import timezone
from django.views import View

from .apis import post_or_put_celery_log
from .helpers import LogSettings, format_django_traceback, get_log_dict

settings: LogSettings = getattr(settings, "LOG_SETTINGS", None)


class TaskLogger(celery.Task):

    def add_log(
        self,
        level: Literal["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        message: str,
    ):
        """
        Add log to the log object
        """
        if not hasattr(self.request, "add_log") and callable(self.request.add_log):
            raise AttributeError(
                "Request object does not have add_log method. Add LogMiddleware to the middleware list"
            )

        self.request.add_log(level, message)

    def __call__(self, *args, **kwargs):
        if not settings.enabled:
            return super().__call__(*args, **kwargs)
        log = {
            "task_id": self.request.id,
            "task_name": self.name,
            "periodic_task_name": (
                self.request.periodic_task_name[:255]
                if hasattr(self.request, "periodic_task_name")
                else None
            ),
            "args": args or None,
            "kwargs": kwargs or None,
            "context": self.request.__dict__,
            "started_at": timezone.now().timestamp(),
        }
        post_or_put_celery_log(log, "POST")
        error = None
        updated_log = {"task_id": self.request.id}
        to_return = None

        def _add_log(level, message):

            log = get_log_dict(level, message)

            if not log:
                return

            if not updated_log.get("logs"):
                updated_log["logs"] = []

            updated_log["logs"].append(log)

        setattr(self.request, "add_log", _add_log)

        try:
            to_return = super().__call__(*args, **kwargs)
        except Exception as e:
            updated_log["error"] = {
                "error": str(e),
                "traceback": traceback.format_exc(),
                "formatted_traceback": format_django_traceback(
                    exc_value=e,
                    exc_tb=e.__traceback__,
                    project_folder=settings.project_folder,
                ),
            }
            error = e
        finally:
            updated_log["ended_at"] = timezone.now().timestamp()
            post_or_put_celery_log(updated_log, "PUT")
            if error:
                raise error
        return to_return


class APILogView(View):
    """
    Extends this class to have code editor suggestions for log settings
    """

    skip_request_body: bool
    skip_request_body_methods: list[str]
    skip_request_headers: bool
    skip_request_headers_methods: list[str]
    skip_response_body: bool
    skip_response_body_methods: list[str]
    priority_log_methods: list[str]

    def add_log(
        self,
        level: Literal["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        message: str,
    ):
        """
        Add log to the log object
        """
        if not hasattr(self.request, "add_log") and callable(self.request.add_log):
            raise AttributeError(
                "Request object does not have add_log method. Add LogMiddleware to the middleware list"
            )

        self.request.add_log(level, message)



import os
import traceback

from asgiref.sync import iscoroutinefunction, markcoroutinefunction
from django.conf import settings
from django.http import HttpRequest
from django.urls import resolve
from django.utils import timezone

from .apis import post_api_log
from .helpers import LogSettings, format_django_traceback, get_log_dict


class LoggingMiddleware:
    async_capable = True
    sync_capable = False

    settings: LogSettings

    method = None
    request_body = None
    headers = None

    skip_request_body = False
    skip_request_body_methods = []
    skip_request_headers = False
    skip_request_headers_methods = []
    skip_response_body = False
    skip_response_body_methods = []
    priority_log_methods = []

    _log = {}

    def __init__(self, get_response):
        self.settings = getattr(settings, "LOG_SETTINGS")
        self.get_response = get_response
        if iscoroutinefunction(self.get_response):
            markcoroutinefunction(self)

    def __is_path_excluded__(self, path: str):
        return any(
            path.startswith(excluded_path)
            for excluded_path in self.settings.paths_to_exclude
        )

    @property
    def __can_skip_logging__body(self):
        return self.skip_request_body or self.method in self.skip_request_body_methods

    @property
    def __can_skip_logging__headers(self):
        return (
            self.skip_request_headers
            or self.method in self.skip_request_headers_methods
        )

    @property
    def __can_skip_logging__response_body(self):
        return self.skip_response_body or self.method in self.skip_response_body_methods

    @property
    def __is_api_priority_log__(self):
        return self.method in self.priority_log_methods

    def update_post_response_data(self, request, response):
        # in django request's user is SimpleLazyObject, if we try to access it before response is composed, it will raise an error since it's not yet evaluated and evaluation will make db call in sync mode
        if (
            not hasattr(request.user, "_wrapped")
            or type(request.user._wrapped) != object
        ):  # if user object has no "_wrapped" attribute or it's not an SimpleLazyObject, then it's already evaluated
            # this way we can check if user is SimpleLazyObject, if request.user is evaluated, it will be of some User model type
            self._log["user"] = {
                "id": request.user.id if request.user.is_authenticated else None,
                "email": (
                    request.user.email if request.user.is_authenticated else None
                ),
            }

        self._log["status_code"] = self.settings.get_status_code(response)
        self._log["headers"] = (
            self.settings.clean_header(dict(request.headers))
            if not self.__can_skip_logging__headers
            else None
        )
        content_type = self._log["headers"].get("Content-Type")
        if content_type != "application/json" or self.__can_skip_logging__body:
            self._log["body"] = (
                None  # body can't be accessed after response is composed
            )
        else:
            if self._log["body"] and isinstance(self._log["body"], bytes):
                self._log["body"] = self._log["body"].decode(
                    "utf-8"
                )  # decode bytes to string

        self._log["response"] = (
            response.content.decode("utf-8")
            if not self.__can_skip_logging__response_body
            and response.headers.get("Content-Type") == "application/json"
            else None
        )
        self._log["ended_at"] = timezone.now().timestamp()

    async def __call__(self, request: HttpRequest):
        if self.__is_path_excluded__(request.path) or not self.settings.enabled:
            return await self.get_response(request)

        view, _, _ = resolve(request.path)
        self.skip_request_body = getattr(
            view,
            "view_class",
            None,
        ) and getattr(
            view.view_class,
            "skip_request_body",
            False,
        )
        self.skip_request_body_methods = (
            getattr(
                view,
                "view_class",
                None,
            )
            and getattr(
                view.view_class,
                "skip_request_body_methods",
                None,
            )
            or []
        )
        self.skip_request_headers = getattr(
            view,
            "view_class",
            None,
        ) and getattr(
            view.view_class,
            "skip_request_headers",
            False,
        )
        self.skip_request_headers_methods = (
            getattr(
                view,
                "view_class",
                None,
            )
            and getattr(
                view.view_class,
                "skip_request_headers_methods",
                None,
            )
            or []
        )
        self.skip_response_body = getattr(
            view,
            "view_class",
            None,
        ) and getattr(
            view.view_class,
            "skip_response_body",
            False,
        )
        self.priority_log_methods = (
            getattr(
                view,
                "view_class",
                None,
            )
            and getattr(
                view.view_class,
                "priority_log_methods",
                None,
            )
            or []
        )

        self._log = {
            "path": request.path,
            "query_params": request.GET.dict(),
            "method": request.method,
            "started_at": timezone.now().timestamp(),
        }

        self._log["body"] = request.body

        def _add_log(level, message):

            log = get_log_dict(level, message)

            if not log:
                return

            if not self._log.get("logs"):
                self._log["logs"] = []

            self._log["logs"].append(log)

        setattr(request, "add_log", _add_log)

        response = await self.get_response(request)
        self.update_post_response_data(request, response)
        post_api_log(self._log, self.__is_api_priority_log__)

        return response

    def process_exception(self, request: HttpRequest, e):
        if not self.settings.enabled:
            return None

        response = self.settings.app_exception_handler(
            request, e, traceback.format_exc()
        )
        if not self.settings.can_ignore_exception(e):
            self._log["error"] = {
                "error": str(e),
                "traceback": traceback.format_exc(),
                "formatted_traceback": format_django_traceback(
                    exc_value=e,
                    exc_tb=e.__traceback__,
                    project_folder=self.settings.project_folder,
                ),
            }
        return response



import requests
from django.conf import settings
from concurrent.futures import ThreadPoolExecutor
from .helpers import LogSettings

settings: LogSettings = getattr(settings, "LOG_SETTINGS", None)

executor = ThreadPoolExecutor(settings.thread_pool_size)


def post_api_log(log_data: dict, priority: bool = False):
    url = (
        settings.url + "/api/api-log/"
        if not settings.url.endswith("/")
        else settings.url + "api/api-log/"
    )
    headers = {
        "content-type": "application/json",
        "x-env-id": settings.env_key,
    }
    query = {"p": "1"} if priority else None

    if settings.use_thread_pool:
        executor.submit(
            requests.post,
            url,
            headers=headers,
            params=query,
            json=log_data,
        )
    else:
        requests.post(
            url,
            headers=headers,
            params=query,
            json=log_data,
        )


def post_or_put_celery_log(log_data: dict, method: str):
    url = (
        settings.url + "/api/celery-log/"
        if not settings.url.endswith("/")
        else settings.url + "api/celery-log/"
    )
    headers = {
        "content-type": "application/json",
        "x-env-id": settings.env_key,
    }

    if settings.use_thread_pool:
        executor.submit(
            requests.request,
            method,
            url,
            headers=headers,
            json=log_data,
        )
    else:
        requests.request(
            method,
            url,
            headers=headers,
            json=log_data,
        )



import inspect
import os
import sys
import threading
import time
import traceback

# import logging
from logging import _nameToLevel, root
from typing import Callable

from asgiref.sync import iscoroutinefunction, markcoroutinefunction
from django.conf import settings
from django.http import HttpRequest, HttpResponse


class LogSettings:
    def __init__(
        self,
        url: str,
        env_key: str,
        enabled: bool,
        app_exception_handler: Callable[[HttpRequest, Exception, str], HttpResponse],
        can_ignore_exception: Callable[[Exception], bool] = lambda x: False,
        paths_to_exclude: list[str] = [],
        clean_header: Callable[[dict], dict] = lambda x: x,
        get_status_code: Callable[[HttpResponse], int] = lambda x: x.status_code,
        use_thread_pool: bool = True,
        thread_pool_size: int = 24,
        project_folder: str = "/backend/",
    ):
        self.url = url
        self.env_key = env_key
        self.enabled = enabled
        self.app_exception_handler = app_exception_handler
        self.can_ignore_exception = can_ignore_exception
        self.paths_to_exclude = paths_to_exclude
        self.clean_header = clean_header
        self.get_status_code = get_status_code
        self.use_thread_pool = use_thread_pool
        self.thread_pool_size = thread_pool_size
        self.project_folder = project_folder


class MiddlewareWrapper:
    """
    This class is a wrapper around the middleware class, it helps to catch exceptions raised by custom middleware and handle them by the app_exception_handler.
    """

    settings: LogSettings
    async_capable = True
    sync_capable = False

    def __init__(self, get_response):
        self.settings = getattr(settings, "LOG_SETTINGS")
        self.get_response = get_response
        if iscoroutinefunction(self.get_response):
            markcoroutinefunction(self)

    async def process(self, request):
        raise NotImplementedError

    async def __call__(self, request):
        try:
            return await self.process(request)
        except Exception as e:
            if not self.settings.enabled or not self.settings.can_ignore_exception(e):
                raise e
            return self.settings.app_exception_handler(
                request, e, traceback.format_exc()
            )


def _format_django_traceback(exc_value, exc_tb, project_folder="/backend/", prefix=""):
    """Format traceback to include both Django and project-specific code with highlights."""
    output = []
    tb = exc_tb  # Start with the traceback object

    while tb:
        frame = tb.tb_frame  # Get the frame object
        lineno = tb.tb_lineno  # Line number where the exception occurred
        filename = frame.f_code.co_filename  # File name of the frame
        funcname = frame.f_code.co_name  # Function name of the frame

        # Check if the frame belongs to the project
        is_project_code = project_folder in filename

        if not is_project_code:
            # move to the next traceback object
            tb = tb.tb_next
            continue

        output.append(
            f"{prefix}File: {filename}, Line: {lineno}, Function: {funcname}\n"
        )
        output.append(
            prefix + filename.replace(project_folder, "") + ":" + str(lineno) + "\n"
        )

        try:
            with open(filename, "r") as f:
                lines = f.readlines()
                context_lines = lines[
                    max(0, lineno - 4) : lineno + 3
                ]  # 3-4 lines around the error
                for i, line in enumerate(context_lines, start=max(0, lineno - 4)):
                    suffix = ""
                    if i + 1 == lineno:
                        suffix = " <--- This caused the exception"

                    if line.startswith("    "):
                        line = line[4:]  # Removing first indentation

                    if line[-1] == "\n":
                        line = line[:-1] + suffix + "\n"
                    else:
                        line = line + suffix
                    output.append(f"{prefix}{i + 1}: {line}")
        except Exception as e:
            output.append(f"{prefix}Could not retrieve surrounding code: {e}\n")

        # Add local variables
        local_vars = frame.f_locals
        if local_vars:
            output.append(f"\n{prefix}Local Variables:\n")
            for var, value in local_vars.items():
                # check if the value are not primitive types
                if not isinstance(value, (int, float, str, bool, list, dict)):
                    continue
                output.append(f"{prefix}     {var}: {value}\n")

        # add a line break
        output.append(
            "\n"
            + prefix
            + "-" * 10
            + f"End of project file {filename}"
            + "-" * 10
            + "\n\n"
        )

        tb = tb.tb_next  # Move to the next traceback object

    if exc_value.__cause__:
        output.append(
            "\nThe above exception happened while handling the following exception:\n"
        )
        cause_tb = exc_value.__cause__.__traceback__
        output.append(
            _format_django_traceback(
                type(exc_value.__cause__),
                cause_tb,
                project_folder,
                prefix + "  ",
            )
        )

    return "".join(output)


def format_django_traceback(exc_value, exc_tb, project_folder="/backend/"):
    try:
        return _format_django_traceback(exc_value, exc_tb, project_folder)
    except Exception as e:
        return f"Error while formatting traceback: \n\nError: {e}\n\ntraceback{traceback.format_exc()}"


def get_log_dict(level: str, message: str):
    current_level = root.getEffectiveLevel()
    level_int = _nameToLevel[level]

    # skip logs with level lower than current level
    if level_int < current_level:
        return None

    frame = (
        inspect.currentframe().f_back.f_back.f_back
    )  # go back 3 frames to get the actual caller
    code = frame.f_code
    filename = code.co_filename
    lineno = frame.f_lineno
    func_name = frame.f_code.co_name

    # Get the time information
    created_time = time.time()
    asctime = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(created_time))
    msecs = int((created_time - int(created_time)) * 1000)
    relative_created = int(created_time * 1000)  # Relative time in milliseconds
    thread_id = threading.get_ident()
    thread_name = threading.current_thread().name
    process_id = os.getpid()
    processName = "MainProcess"
    mp = sys.modules.get("multiprocessing")
    if mp is not None:
        # Errors may occur if multiprocessing has not finished loading
        # yet - e.g. if a custom import hook causes third-party code
        # to run when multiprocessing calls import. See issue 8200
        # for an example
        try:
            processName = mp.current_process().name
        except Exception:  # pragma: no cover
            pass

    return {
        "levelname": level,
        "pathname": filename,  # Full pathname of the source file
        "filename": os.path.basename(filename),  # Filename portion of the pathname
        "module": filename.split("/")[-1].split(".")[0],  # Module name from filename
        "lineno": lineno,  # Line number where log is generated
        "funcName": func_name,  # Function name
        "created": created_time,  # Time when the LogRecord was created (in seconds)
        "asctime": asctime,  # Textual time when the LogRecord was created
        "msecs": msecs,  # Millisecond portion of the creation time
        "relativeCreated": relative_created,  # Relative time in milliseconds since app start
        "thread": thread_id,
        "threadName": thread_name,
        "process": process_id,
        "processName": processName,
        "message": message,
    }
